#------------------------------------------------------------------
# Contact: Shawn Patton
# Created: Jan 2009
#
# Purpose: Handsome Rob's Daisy Chain
#
#------------------------------------------------------------------

import random
import os.path
import sys
import logging
import traceback

import pygame
from pygame.locals import *

from core import util

from core.Player import Player
from core.Flower import Flower
from core.Ground import Ground
from core.Narrator import Narrator
from core.Tutorial import Tutorial
from core.EventLog import EventLog
from core.InputManager import InputManager
from core.InputManager import KeyData
from core.WelcomeScreen import WelcomeScreen
from core.FarewellScreen import FarewellScreen
from core.Camera import Camera

from net.HandsomeClient import HandsomeClient

# each type of game object gets an init and an
# update function. the update function is called
# once per frame, and it is when each object should
# change its current position and state. the Player
# object actually gets a "move" function instead of
# update, since it is passed extra information about
# the keyboard

class DaisyChain:
	logger = logging.getLogger( "DaisyChain" )
	logger.setLevel( logging.DEBUG )
	
	def __init__(self):
		self.initVariables()
		self.initLogger()
		self.initPygame()
		self.loadImages()
		self.initWindow()
		self.initGameData()

	def initVariables(self):
		self.clock = pygame.time.Clock()
		#define the input keys
		#convenience array of keys for generic operations
		#only put unique keys in here
		
		self.playerInput = InputManager()
		self.recordedInput = InputManager()
		self.starterFlowerPositions = [50, 120, 156, 220, 270, 380, 390, 412, 458, 480, 530, 589, 612, 640, 680, 720, 758, 770 ]
		self.gameMode = "welcome"  # welcome, play, or farewell

		self.volume = 0.3
		self.pollenFactor = 2
		self.maxFlowers = 25
		self.relativePollenDestinationXs = [-2,25]
		self.minDistanceBetweenFlowers = 25
		self.tickCount = 0
		self.timeLeftToSimulate = 0   # in milliseconds
		self.stepDT = 10  # in milliseconds
		self.stepDTInSeconds = self.stepDT * 0.001
		self.wasUpDown = False
		self.wasDownDown = False
		self.flowerBaseValue = 10
		self.pollinateBonus = 10

		self.gameLength = 75*1000  # in milliseconds
		self.currentGameTime = 0  # in milliseconds
		
	def initLogger( self ):
		logging.basicConfig( format="%(name)s: %(message)s" )

	def initPygame(self):
		# Initialize pygame
		pygame.init()
		if pygame.mixer and not pygame.mixer.get_init():
			DaisyChain.logger.warning( "No sound support" )
			pygame.mixer = None
		# Set the display mode
		winstyle = 0  # |FULLSCREEN
		self.screenRect = Rect(0, 0, 800, 600)
		bestdepth = pygame.display.mode_ok(self.screenRect.size, winstyle, 32)
		self.screen = pygame.display.set_mode(self.screenRect.size, winstyle, bestdepth)

	def initWindow(self):
		# decorate the game window
#		icon = pygame.transform.scale(self.daisyImage, (32, 32))
#		pygame.display.set_icon(icon)
		pygame.display.set_caption( "Daisy Chain" )
		pygame.mouse.set_visible( 0 )

		# Sprite groups
		self.allSprites = pygame.sprite.Group()
		self.groundSprites = pygame.sprite.Group()
		self.playerSprites = pygame.sprite.Group()
		self.flowerSprites = pygame.sprite.Group()
	
	def initGameData(self):
		# Init game stuff
		self.welcomeScreen = WelcomeScreen()
		self.farewellScreen = FarewellScreen()
		self.camera = Camera(self, self.screenRect.width, self.screenRect.height)
		self.camera.setPercent( 0.0 )
		# set camera timeline as a percentage of gameLength
		self.cameraTimeline = [(0.00*self.gameLength, 0.0),
		                       (0.20*self.gameLength, 0.0),
		                       (0.23*self.gameLength, 0.4),
		                       (0.66*self.gameLength, 0.4),
		                       (0.70*self.gameLength, 1.0),
		                       (1.00*self.gameLength, 1.0)]

		self.topGround = Ground(0, self,  [self.allSprites, self.groundSprites], 0, None, [self.allSprites, self.flowerSprites])
		self.middleGround = Ground(1, self, [self.allSprites, self.groundSprites], self.screenRect.height/2, self.topGround, [self.allSprites, self.flowerSprites])
		self.bottomGround = Ground(2, self, [self.allSprites, self.groundSprites], self.screenRect.height, self.middleGround, [self.allSprites, self.flowerSprites]) 
		
		# initialize with floor, ceiling
		# playerPlayer in on the top
		self.playerPlayer = Player( self, "Handsome Rob", [self.allSprites, self.playerSprites], self.middleGround, self.topGround )
		# recorded player is on the bottom
		self.recordedPlayer = Player( self, "Handsome Rob's Former Handsome Self", [self.allSprites, self.playerSprites], self.bottomGround, self.middleGround, isLive=False )
		
		self.playerEventLog = EventLog("Player")
		
		# connect to the server
		
		self.recordedEventLog = EventLog("Handsome Rob")
		
		try:
			self.handsomeClient = HandsomeClient()
			data = self.handsomeClient.getGameData()
			self.recordedEventLog.initFromData( data )
			if self.recordedEventLog.failedLoad:
				print "Server data failed; running from local data"
				self.recordedEventLog.readLog("lastRun.log")
				if self.recordedEventLog.failedLoad:
					DaisyChain.logger.info( "Load from local file failed" )
		except:
			traceback.print_exc()
			print "Server connection failed; running from local data"
			self.recordedEventLog.readLog("lastRun.log")
			if self.recordedEventLog.failedLoad:
				DaisyChain.logger.info( "Load from local file failed" )
		
		self.recordedPlayer.setName(self.recordedEventLog.playerName)

		# "seed" with a few flowers... heheheh.. seed!
		chosenStarterFlowerPositions = random.sample( self.starterFlowerPositions, 5 )
		for starterFlowerPosition in chosenStarterFlowerPositions:
			self.middleGround.growFlower( starterFlowerPosition, Flower.BLOOM )	
			
		# initialize Narrator
		self.narrator = Narrator(self)
		
		# initialize Tutorial
		self.tutorial = Tutorial(self)
	
	def loadImages( self ):
		# Load images, assign to sprite classes
		# (do this before the classes are used, after screen setup)

		# create the background
		self.background = util.load_image('assets/textures/background.jpg')

	def run( self ):
		# Main game loop
		while True:
			dt = self.clock.tick( 60 )

			# Handle input
			for event in pygame.event.get():
				if event.type == QUIT:
					self.exit()
				elif self.gameMode == "welcome":
					if event.type == KEYDOWN:
						if event.key == pygame.K_RETURN:
							pygame.mixer.music.fadeout( 500 )
							pygame.mixer.music.load( os.path.join( "assets/music/", "game.mp3" ) )
							pygame.mixer.music.play(2)
							pygame.mixer.music.set_volume( self.volume )
							self.screen.blit( self.background, (0, 0) )
							pygame.display.flip()
							self.playerPlayer.setName(self.welcomeScreen.name)
							self.playerEventLog.playerName = self.playerPlayer.name
							self.gameMode = "play"
						else:
							self.welcomeScreen.handleKeyEvent( event )
				elif self.gameMode == "play":
					pass
				elif self.gameMode == "farewell":
					pass

			# Simulation
			if self.gameMode == "play":
				if self.currentGameTime < self.gameLength:
					self.simulate( dt )
					self.allSprites.update()
					self.narrator.update()
					self.tutorial.update()
				else:
					self.gameMode = "farewell"
					self.farewellScreen.startFarewell( self.playerPlayer.flowersPicked, self.playerPlayer.flowersPulled )
					try:
						self.handsomeClient.postGameData( self.playerEventLog.getData() )
					except:
						traceback.print_exc()
						print "Failed to post game data to server."
			elif self.gameMode == "farewell":
				self.farewellScreen.update()

			# Rendering
			if self.gameMode == "welcome":
				self.welcomeScreen.render( self.screen )
			elif self.gameMode == "play":
				# Draw background
				self.screen.blit( self.background, (0, self.camera.getCameraHeightOffset()) )

				# Draw all sprites
				self.groundSprites.draw( self.screen )
				
				self.playerSprites.draw( self.screen )
				for player in self.playerSprites:
					player.renderOtherHalf( self.screen )
				self.flowerSprites.draw( self.screen )
				self.recordedPlayer.particleGroup.draw( self.screen )
				self.playerPlayer.particleGroup.draw( self.screen )

				# Draw the name tags
				self.playerPlayer.nameTag.render(self.screen)
				self.recordedPlayer.nameTag.render(self.screen)
				# Draw black bars
				self.camera.drawLetterBoxes()
				# draw tutorial
				self.tutorial.draw( self.screen )
			elif self.gameMode == "farewell":
				self.farewellScreen.render( self.screen )
			pygame.display.flip()
	
	def simulate( self, dt ):
		self.timeLeftToSimulate += dt
		# Do any per-frame pre-simulation housekeeping.

		# Simulate
		while self.timeLeftToSimulate >= self.stepDT:
			self.currentGameTime += self.stepDT
			if self.currentGameTime < self.cameraTimeline[0][0]:
				percent = self.cameraTimeline[0][1]
			elif self.currentGameTime >= self.cameraTimeline[-1][0]:
				percent = self.cameraTimeline[-1][1]
			else:
				for i in range( len( self.cameraTimeline ) - 1 ):
					millis1, percent1 = self.cameraTimeline[i]
					millis2, percent2 = self.cameraTimeline[i + 1]
					if (self.currentGameTime >= millis1) and (self.currentGameTime < millis2):
						percent = util.interpolate( percent1, percent2, float( self.currentGameTime - millis1 )/float( millis2 - millis1 ) )
			self.camera.setPercent( percent )

			keyState = pygame.key.get_pressed()
			# get input from keyboard
			# record the key events
			for key in self.playerInput.uniqueKeys:
				isDown = keyState[key.keyValue]
				if (isDown and not(key.wasDown )):
					self.recordEvent(EventLog.KEY_DOWN, [key.keyValue])
				elif (not(isDown) and key.wasDown):
					self.recordEvent(EventLog.KEY_UP, [key.keyValue])
			# set the key state
			for key in self.playerInput.allKeys:
				key.setIsDown(keyState[key.keyValue])
			
			# simulate the input controlled player 
			self.simulateStep( self.playerPlayer, self.playerInput )

			# Read the recorded input and see if it is time to do something
			while (self.recordedEventLog.getTopEventTickTime() == self.tickCount):
				currentRecordedEvent = self.recordedEventLog.getTopEvent()
				
				if (currentRecordedEvent[1] == EventLog.KEY_DOWN):
					self.recordedInput.setKeyState(currentRecordedEvent[2][0], True)
				if (currentRecordedEvent[1] == EventLog.KEY_UP):
					self.recordedInput.setKeyState(currentRecordedEvent[2][0], False)
				if (currentRecordedEvent[1] == EventLog.FLOWER_GREW):
					# level, x
					# determine which ground this event is for
					
#					if currentRecordedEvent[2][0] == 0: # previous player's top ground
#						self.middleGround.growFlower(currentRecordedEvent[2][1], Flower.SEED)
					if currentRecordedEvent[2][0] == 1: # previous player's middle ground
						DaisyChain.logger.debug( "prerecorded flower grew event for ground %d" % currentRecordedEvent[2][0] )
						self.bottomGround.growFlower(currentRecordedEvent[2][1], Flower.SEED)
						

						
#				if (currentRecordedEvent[1] == EventLog.FLOWER_PULLED):
#					if currentRecordedEvent[2][0] == 0: # previous player's top ground
#						self.middleGround.pullFlower(currentRecordedEvent[2][1])
#					if currentRecordedEvent[2][0] == 1: # previous player's middle ground
#						self.bottomGround.pullFlower(currentRecordedEvent[2][1])
				# remove that event from the event log
				self.recordedEventLog.removeTopEvent()

			# simulate the recorded player 
			self.simulateStep( self.recordedPlayer,  self.recordedInput )
			
			# simulate the ground
			self.topGround.simulateStep(self.stepDTInSeconds)
			self.middleGround.simulateStep(self.stepDTInSeconds)
			self.bottomGround.simulateStep(self.stepDTInSeconds)
			
			
			self.timeLeftToSimulate -= self.stepDT
			self.tickCount = self.tickCount + 1

		# Do any per-frame post-simulation housekeeping.

	def simulateStep( self, whichPlayer, inputState ):
		# Alter game state based on keystate and self.stepDT.
		
		if (inputState.RIGHT_KEY.isDown) :
			whichPlayer.setMovingRight(True)
		else:
			whichPlayer.setMovingRight(False)
			
		if (inputState.LEFT_KEY.isDown):
			whichPlayer.setMovingLeft(True)
		else:
			whichPlayer.setMovingLeft(False)
			
		if (inputState.JUMP_KEY.isDown and not(inputState.JUMP_KEY.wasDown)):
			whichPlayer.jump()
		if (inputState.PICK_KEY.isDown and not(inputState.PICK_KEY.wasDown)):
			whichPlayer.pick()
		if (inputState.PULL_KEY.isDown and not(inputState.PULL_KEY.wasDown)):
			whichPlayer.pull()
			
		whichPlayer.simulateStep( self.stepDTInSeconds )
		
	def recordEvent(self, eventName, eventData):
		self.playerEventLog.addEvent(self.tickCount, eventName, eventData)
		
	def exit(self):
		self.playerEventLog.writeLog("lastRun.log")
		sys.exit()	
