#-------------------------------------------------------------------------------
# Contact: Rob Gordon
# Created: Jan 2009
#
# Purpose: Ground layer that handles flowers in it.
#-------------------------------------------------------------------------------

# Python imports
import logging
import random
import math

from core import util
#import basic pygame modules
import pygame
from core.Flower import Flower
from core.Pollen import Pollen
from core.EventLog import EventLog

from pygame.locals import *

class Ground(pygame.sprite.Sprite):

	logger = logging.getLogger( "Ground" )
	logger.setLevel( logging.INFO )

	def __init__(self, id, daisyChain, groups, height, ceiling, flowerGroups):
		pygame.sprite.Sprite.__init__(self, groups)
		self.id = id
		Ground.logger.debug( "ground ID: %d" % id )
		self.daisyChain = daisyChain
		self.ceiling = ceiling
		self.flowerGroups = flowerGroups
		self.flowerIdsToFlowers = {}
		self.pollenIdsToPollen = {}		
		self.height = height # y position of this ground in world space
		self.thickness = 48 # how thick (in y) the ground is. Used for
		                    # determining flower stem and root positions
		self.dataTop = self.height - self.thickness/2
		self.dataBottom = self.height - self.thickness/2

		self.image = util.load_image('assets/textures/ground/ground.png').convert_alpha()
		centerPos = (self.daisyChain.screenRect.center[0], self.height)
		self.rect = self.image.get_rect(center=centerPos)
		self.setImagePositionBasedOnData()

	def releasePollen(self, x):
		Ground.logger.debug( "Ground %d releasing pollen" % self.id )
		periods = [(random.random()*2.0 + 2.0) * math.pi, -1.0 * (random.random()*3.0 + 1.0) * math.pi]

		# create pollen from picked flower
		if self.ceiling != None: 
			# first make sure this doesn't put us over the range of flowers possible
			# and make sure this slot isn't full
			periodIndex = 0
			for destX in self.daisyChain.relativePollenDestinationXs:
				period = periods[periodIndex]
				newX = x + destX
				if ( len(self.flowerIdsToFlowers.keys()) < self.daisyChain.maxFlowers ) and self.ceiling.pollenDestinationIsFree(newX) :					
					Ground.logger.debug( "adding pollen to ground %d at position %.1f" % (self.ceiling.id, newX) )
					self.addPollen( Pollen( self.daisyChain, self.flowerGroups, self, self.ceiling, x, newX, period) )

	def pollenDestinationIsFree( self, goalX ):		
		# check all existing pollen and flower
		if self.ceiling != None: 
			for flower in self.ceiling.flowerIdsToFlowers.values():
				if goalX < (flower.x + self.daisyChain.minDistanceBetweenFlowers) and goalX > (flower.x - self.daisyChain.minDistanceBetweenFlowers):
					flower.pollinate()
					return False	
		for pollen in self.pollenIdsToPollen.values():
			
			if goalX < (pollen.destX + self.daisyChain.minDistanceBetweenFlowers) and goalX > (pollen.destX - self.daisyChain.minDistanceBetweenFlowers):
				return False
		return True
			
	def addPollen( self, pollen ):
		self.pollenIdsToPollen[pollen.id] = pollen

	def removePollen( self, pollen ):
		if self.pollenIdsToPollen.has_key(pollen.id):
			del self.pollenIdsToPollen[pollen.id]

	def growFlower( self, x, startState ):
		self.daisyChain.recordEvent(EventLog.FLOWER_GREW, [self.id, int(x)])
		Ground.logger.debug( "growing flower on ground %d at position %.1f" % (self.id, x) )
		self.addFlower(	Flower( self.daisyChain, self.flowerGroups, self, x, startState ) )

	def addFlower( self, flower ):
		self.flowerIdsToFlowers[flower.id] = flower

	def removeFlower( self, flower ):
		if self.flowerIdsToFlowers.has_key(flower.id):
			del self.flowerIdsToFlowers[flower.id]
		
	def getImageSpaceHeight(self):
		return self.height + self.daisyChain.camera.getCameraHeightOffset()

	def setImagePositionBasedOnData(self):
		self.rect.centery = self.getImageSpaceHeight()
		
	def simulateStep( self, dT ):
		self.setImagePositionBasedOnData()
		# step all flowers
		for flower in self.flowerIdsToFlowers.values():
			flower.simulateStep(dT)
		for pollen in self.pollenIdsToPollen.values():
			pollen.simulateStep(dT)
