#------------------------------------------------------------------
# Contact: Shawn Patton
# Created: Jan 2009
#
# Purpose: The player who picks flowers.
#------------------------------------------------------------------

# Python imports
import os
import math
import logging
import pygame
import random
from pygame.locals import *

# Daisy Chain imports
from core import util
from core.Flower import Flower
from core.Animator import Animator
from core.NameTag import NameTag
from core.Particle import Particle

class Player(pygame.sprite.Sprite):
	speed = 150 # units/second
	images = []
	jumpSpeed = -6.7 # units/second
	GRAVITY = -9.8 * 1.5
	pickingThreshold = 50 # the Player's x position needs to be this close to
	                      # a Flower's x position for a Pick to go through
	pullingThreshold = 50 # the Player's x position needs to be this close to
	                       # a Flower's x position for a Pull to go through
	maxFlowerDistance = 99999
	pullingHeight = 90.0 # player must be at least this high off the floor for a
	                      # Pull to go through.
	positionX = 400
	positionY = 0
	
	SADNESS_TIME = 3
	sadnessCounter = 0
	
	HAPPINESS_TIME = 3
	happinessCounter = 0
	
	particles = []
	
	logger = logging.getLogger( "Player" )
	logger.setLevel( logging.INFO )
	fsmLogger = logging.getLogger( "PlayerFSM" )
	fsmLogger.setLevel( logging.INFO )
	
	def __init__( self, daisyChain, name, groups, floor, ceiling, isLive=True ):
		self.daisyChain = daisyChain
		self.name = name
		self.floor = floor # the Ground immediately below this player
		self.ceiling = ceiling # the Ground immediately above this player
		self.flowersPickedCount = 0
		self.sound = None
		self.isLive = isLive
		self.initAudio()
		self.flowersPicked = 0
		self.flowersPulled = 0
		# valid states are:
		# -Picking
		# -Pulling
		# -Walking
		# -Idle
		# -OffGround
		self.state = "Idle"
		self.pickingAnimationDone = False
		self.pullingAnimationDone = False
		self.pickRequested = False
		self.jumpRequested = False
		self.pullRequested = False
		self.movingLeft = False
		self.movingRight = False
		self.happyActionCount = 0
		self.evilActionCount = 0
		self.facing = -1 # -1 for left, 1 for right
		self.playerToFlowerDirection = -1
		# drawing stuff
		pygame.sprite.Sprite.__init__(self, groups)

		self.particleGroup = pygame.sprite.Group()
		self.animators = {}
		self.animators["Idle"] = Animator(self, "player/player_idle_%d.png", numberOfFrames=8, fps=6)
		self.animators["WalkLeft"] = Animator(self, "player/player_walkLeft_%d.png", numberOfFrames=4, fps=6)
		self.animators["WalkRight"] = Animator(self, "player/player_walkRight_%d.png", numberOfFrames=4, fps=6)
		self.animators["AirIdleRight"] = Animator(self, "player/player_airIdleRight_%d.png", numberOfFrames=3, fps=8)
		self.animators["AirIdleLeft"] = Animator(self, "player/player_airIdleLeft_%d.png", numberOfFrames=3, fps=8)
		self.animators["PickRight"] = Animator(self, "player/player_pickRight_%d.png", numberOfFrames=6, fps=12)
		self.animators["PickLeft"] = Animator(self, "player/player_pickLeft_%d.png", numberOfFrames=6, fps=12)
		self.animators["PullRight"] = Animator(self, "player/player_pullRight_%d.png", numberOfFrames=4, fps=9)
		self.animators["PullLeft"] = Animator(self, "player/player_pullLeft_%d.png", numberOfFrames=4, fps=9)
		
		self.image = self.animators["Idle"].images[0]
		self.positionX = self.daisyChain.screenRect.centerx
		self.positionY = self.floor.dataTop
		startPos = (self.daisyChain.screenRect.center[0], self.floor.rect.top)
		self.rect = self.image.get_rect(midbottom=startPos)
		self.setImagePositionBasedOnDataPosition()
	
	def initAudio( self ):
		self.baseSounds = {
			"jump" :  pygame.mixer.Sound(os.path.join("assets/audio/player/","jump.wav")),
			"pickFlower" :  pygame.mixer.Sound(os.path.join("assets/audio/player/","pick.wav")),
			"pullRoot" :  pygame.mixer.Sound(os.path.join("assets/audio/player/","pull.wav")),
		}
		
		self.happySounds = {
			"happy_0" :  pygame.mixer.Sound(os.path.join("assets/audio/player/","happy_0.wav")),
			"happy_1" :  pygame.mixer.Sound(os.path.join("assets/audio/player/","happy_1.wav")),
			"happy_2" :  pygame.mixer.Sound(os.path.join("assets/audio/player/","happy_2.wav")),
			"happy_3" :  pygame.mixer.Sound(os.path.join("assets/audio/player/","happy_3.wav")),
			"happy_4" :  pygame.mixer.Sound(os.path.join("assets/audio/player/","happy_4.wav")),
			"happy_5" :  pygame.mixer.Sound(os.path.join("assets/audio/player/","happy_5.wav")),
			"happy_6" :  pygame.mixer.Sound(os.path.join("assets/audio/player/","happy_6.wav")),
		}
		
		self.evilSounds = {
			"evil_0" :  pygame.mixer.Sound(os.path.join("assets/audio/player/","evil_0.wav")),
			"evil_1" :  pygame.mixer.Sound(os.path.join("assets/audio/player/","evil_1.wav")),
			"evil_2" :  pygame.mixer.Sound(os.path.join("assets/audio/player/","evil_2.wav")),
			"evil_3" :  pygame.mixer.Sound(os.path.join("assets/audio/player/","evil_3.wav")),
			"evil_4" :  pygame.mixer.Sound(os.path.join("assets/audio/player/","evil_4.wav")),
		}
					
	# update player movement state based on input
	def pick( self ):
		self.pickRequested = True
		
	def jump( self ):
		self.jumpRequested = True
	
	def pull( self ):
		self.pullRequested = True
	
	def setMovingLeft(self, moving):
		self.movingLeft = moving
		
	def setMovingRight(self, moving):
		self.movingRight = moving

	def setName(self, name):
		self.name = name
		self.nameTag = NameTag(self, self.name)

	# update player position/action state base on movement state
	def simulateStep(self, dt):
		"""
		dt: duration of the update in seconds
		"""
		self.checkForStateTransitions( dt )
		
		# update player based on current state, dt, and left/right movement keys
		if self.state == "Idle":
			self.animators[self.state].animate( dt )
		elif self.state == "Walking":
			moving = self.updateFacing()
			if moving:
				self.moveSideways( dt )
			# update facing animation as needed
			if self.facing == -1:
				self.animators["WalkLeft"].animate( dt )
			elif self.facing == 1:
				self.animators["WalkRight"].animate( dt )
		elif self.state == "Picking":
			if self.facing == -1:			
				self.animators["PickLeft"].animate(dt, setattr, (self, "pickingAnimationDone", True))
			elif self.facing == 1:
				self.animators["PickRight"].animate(dt, setattr, (self, "pickingAnimationDone", True))
		elif self.state == "OffGround":
			# apply gravity to the vertical speed
			self.verticalSpeed = self.verticalSpeed - (self.GRAVITY * dt)
			self.positionY += self.verticalSpeed
#			topOfPlayer = self.getTopOfPlayerInScreenSpace() 
#			bottomOfLetterBox = self.daisyChain.camera.getTopRect().bottom
#			if (topOfPlayer < bottomOfLetterBox):
#				self.positionY += bottomOfLetterBox - topOfPlayer
#				self.verticalSpeed *= -1
			moving = self.updateFacing()
			if moving:
				self.moveSideways( dt )
			# TODO: update with proper flying animations once we have them
			# update facing animation as needed
			if self.facing == -1:
				self.animators["AirIdleLeft"].animate(dt)
			elif self.facing == 1:
				self.animators["AirIdleRight"].animate(dt)
				
		elif self.state == "Pulling":
			if self.facing == -1:			
				self.animators["PullLeft"].animate(dt, setattr, (self, "pullingAnimationDone", True))
			elif self.facing == 1:
				self.animators["PullRight"].animate(dt, setattr, (self, "pullingAnimationDone", True))

		# Do sadness
		flowerCount = 0
		for flower in self.floor.flowerIdsToFlowers.values():
			if flower.isPickable():
				flowerCount += 1
		if (flowerCount == 0):
			self.sadnessCounter += dt
			if (self.sadnessCounter >= self.SADNESS_TIME):
				self.letSadnessFillTheWorld()
				self.sadnessCounter = 0
		else:
			self.happinessCounter += dt
			if (self.happinessCounter >= self.HAPPINESS_TIME):
				self.happinessFillsMyHeart()
				self.happinessCounter = 0

		# Update any and all particle effects
		self.updateParticles(dt)
			
		
		# clear request flags
		self.pickRequested = False
		self.jumpRequested = False
		self.pullRequested = False

	def updateParticles(self, dt):
		for particleGroup in self.particles:
			for particle in particleGroup:
				particle.simulateStep(dt)
				if (particle.isDone()):
					particleGroup.remove(particle)
			if (len(particleGroup) == 0):
				self.particles.remove(particleGroup)

	def checkForStateTransitions( self, dt ):
		"""
		Change state as need.
		"""
		if self.state == "Idle":
			# order of these checks determines order of state precedence
			if self.pickRequested:
				pickableFlower = self.getPickableFlower()
				if pickableFlower is not None:
					Player.logger.debug( "Player '%s' succeeded in picking flower %d" % (self.name, pickableFlower.id) )
					self.exitIdle()
					self.enterPicking( pickableFlower )
				else:
					Player.logger.debug( "Player %s failed to pick a flower" % self.name )
					# TODO: play a Pick failed animation or some audio feedback
			elif self.jumpRequested:
				self.exitIdle()
				self.enterOffGround()
			# holding down both left/right movement keys cancels left/right movement
			elif self.movingLeft and not self.movingRight:
				self.exitIdle()
				self.enterWalking()
			elif not self.movingLeft and self.movingRight:
				self.exitIdle()
				self.enterWalking()
		elif self.state == "Walking":
			if self.pickRequested:
				pickableFlower = self.getPickableFlower()
				if pickableFlower is not None:
					Player.logger.debug( "Player '%s' succeeded in picking flower %d" % (self.name, pickableFlower.id) )
					self.exitWalking()
					self.enterPicking( pickableFlower )
				else:
					Player.logger.debug( "Player %s failed to pick a flower" % self.name )
					# TODO: play a Pick failed animation or some audio feedback
			elif self.jumpRequested:
				self.exitWalking()
				self.enterOffGround()
			elif not self.movingLeft and not self.movingRight:
				self.exitWalking()
				self.enterIdle()
		elif self.state == "Picking":
			if self.pickingAnimationDone:
				self.pickingAnimationDone = False
				self.exitPicking()
				if self.movingLeft and not self.movingRight:
					self.enterWalking()
				elif not self.movingLeft and self.movingRight:
					self.enterWalking()
				else:
					self.enterIdle()
		elif self.state == "OffGround":
			if (self.positionY > self.floor.dataTop ):
				self.positionY = self.floor.dataTop
				self.exitOffGround()
				if self.movingLeft and not self.movingRight:
					self.enterWalking()
				elif not self.movingLeft and self.movingRight:
					self.enterWalking()
				else:
					self.enterIdle()
			elif self.pullRequested: # see if they are trying to pull a flower
				pullableFlower = self.getPullableFlower()
				if pullableFlower is not None:
					Player.logger.debug( "Player '%s' succeeded in pulling flower %d" % (self.name, pullableFlower.id) )
					self.exitOffGround()
					self.enterPulling( pullableFlower )
				else:
					Player.logger.debug( "Player %s failed to pull a flower" % self.name )
					# TODO: play a Pull failed animation or some audio feedback
		elif self.state == "Pulling":
			if self.pullingAnimationDone:
				self.pullingAnimationDone = False
				self.exitPulling()
				self.enterOffGround( midJump=True )
		else:
			Player.logger.error( "ERROR: Player '%s' state '%s' is invalid." % (self.name, self.state) )

	def enterIdle( self ):
		Player.fsmLogger.debug( "Player '%s' entered Idle" % self.name )
		self.state = "Idle"
				
	def exitIdle( self ):
		pass
		
	def enterWalking( self ):
		Player.fsmLogger.debug( "Player '%s' entered Walking" % self.name )
		self.state = "Walking"
				
	def exitWalking( self ):
		pass
		
		
	def showerLoveAllOverThePlace(self):
		if self.daisyChain.recordedPlayer == self:
			return
		newParticles = []
		startTime = 0
		for dir in [(-75, -30), (-50, -30), (-25, -40), (0, -40), (25, -40), (50, -30), (75, -30)]:
			playerCenter = (self.positionX, self.positionY - self.rect.height/2)
			#animatorInfo = ("player/emotes/emotion_love_%d.png", 2, 8)
			imageFile = "player/emotes/emotion_love.png"
			xMultiplier = random.random()*.5 + .7
			yMultiplier = random.random()*.5 + .6
			intitialDir = (xMultiplier*dir[0], yMultiplier*dir[1])
			startTime = random.random()
			particle = Particle(self.daisyChain, [self.daisyChain.allSprites, self.particleGroup], self, imageFile, playerCenter, intitialDir, 5, startTime)
			newParticles.append(particle)
		self.particles.append(newParticles)
		
	def happinessFillsMyHeart(self):
		if self.daisyChain.recordedPlayer == self:
			return
		newParticles = []
		startTime = 0
		for dir in [(-20, -20), (5, -35), (-10, -35), (30, -25)]:
			playerCenter = (self.positionX, self.positionY - self.rect.height*.75)
			#animatorInfo = ("player/emotes/emotion_love_%d.png", 2, 8)
			imageFile = "player/emotes/emotion_happy.png"
			xMultiplier = random.random()*.5 + .7
			yMultiplier = random.random()*.5 + .6
			intitialDir = (xMultiplier*dir[0], yMultiplier*dir[1])
			particle = Particle(self.daisyChain, [self.daisyChain.allSprites, self.particleGroup], self, imageFile, playerCenter, intitialDir, 3, 0)
			particle.FRICTION_AMOUNT = .1
			newParticles.append(particle)
		self.particles.append(newParticles)
		
	def letSadnessFillTheWorld(self):
		if self.daisyChain.recordedPlayer == self:
			return
		newParticles = []
		startTime = 0
		for dir in [(-20, -20), (5, -35), (-10, -35), (30, -25)]:
			playerCenter = (self.positionX, self.positionY - self.rect.height*.75)
			#animatorInfo = ("player/emotes/emotion_love_%d.png", 2, 8)
			imageFile = "player/emotes/emotion_sad.png"
			xMultiplier = random.random()*.5 + .7
			yMultiplier = random.random()*.5 + .6
			intitialDir = (xMultiplier*dir[0], yMultiplier*dir[1])
			particle = Particle(self.daisyChain, [self.daisyChain.allSprites, self.particleGroup], self, imageFile, playerCenter, intitialDir, 5, 0)
			particle.FLOAT_FORCE = 20
			newParticles.append(particle)
		self.particles.append(newParticles)
	
	def explodeRageEverywhere(self):
		if self.daisyChain.recordedPlayer == self:
			return
		newParticles = []
		startTime = 0
		for dir in [(-75, 20), (-50, 30), (-25, 40), (0, 45), (25, 40), (50, 30), (75, 20)]:
			playerCenter = (self.positionX, self.positionY - self.rect.height*.75)
			#animatorInfo = ("player/emotes/emotion_love_%d.png", 2, 8)
			imageFile = "player/emotes/emotion_angry.png"
			xMultiplier = random.random()*.2 + 1
			yMultiplier = random.random()*.2 + 1
			intitialDir = (xMultiplier*dir[0], yMultiplier*dir[1])
			particle = Particle(self.daisyChain, [self.daisyChain.allSprites, self.particleGroup], self, imageFile, playerCenter, intitialDir, 3, 0)
			particle.FLOAT_FORCE = 1
			particle.FRICTION_AMOUNT = 0
			newParticles.append(particle)
		self.particles.append(newParticles)
		
	def enterPicking( self, flower ):
		Player.fsmLogger.debug( "Player '%s' is Picking flower %d" % (self.name, flower.id) )
		self.state = "Picking"
		self.facing = self.playerToFlowerDirection
		self.happyActionCount += 1
		if self.isLive:
#			self.baseSounds["pickFlower"].play() # TODO: turn this back on if we get a meaningful sound
			happySoundIndex = self.happyActionCount % len( self.happySounds )
			self.happySounds["happy_%d" % happySoundIndex].play()
		self.animators["PickLeft"].reset()
		self.animators["PickRight"].reset()
		
		self.showerLoveAllOverThePlace()
		# tell to flower it is being picked		
		flower.pick()
		self.flowersPicked += 1		
		# TODO: setup so we transition out of Picking when animation is done
				
	def exitPicking( self ):
		pass
		
	def enterOffGround( self, midJump=False ):
		Player.fsmLogger.debug( "Player '%s' entered OffGround" % self.name )
		self.state = "OffGround"
		if not midJump:
			if self.isLive:
				pass
#				self.baseSounds["jump"].play() # TODO: add back in if we have a non-temp sound
			self.verticalSpeed = Player.jumpSpeed
		
	def exitOffGround( self ):
		pass
		
	def enterPulling( self, flower ):
		Player.fsmLogger.debug( "Player '%s' is Pulling flower %d" % (self.name, flower.id) )
		self.state = "Pulling"
		self.animators["PullLeft"].reset()
		self.animators["PullRight"].reset()
		self.facing = self.playerToFlowerDirection
		self.evilActionCount += 1
		if self.isLive:
#			self.baseSounds["pullRoot"].play() # TODO: turn this back on if we get a meaningful sound
			evilSoundIndex = self.evilActionCount % len( self.evilSounds )
			self.evilSounds["evil_%d" % evilSoundIndex].play()
			self.explodeRageEverywhere()
		flower.setState( Flower.DEAD )
		self.flowersPulled += 1
		# TODO: start pulling animation
				
	def exitPulling( self ):
		pass
	
	def getTopOfPlayerInScreenSpace(self):
		return self.positionY - self.rect.height + self.daisyChain.camera.getCameraHeightOffset() + 20
	
	
	def setImagePositionBasedOnDataPosition(self):
		self.rect.bottom = self.positionY + self.daisyChain.camera.getCameraHeightOffset()
		self.rect.centerx = self.positionX
	
	def update( self ):
		self.setImagePositionBasedOnDataPosition()

	def updateFacing( self ):
		"""
		Returns True if the player is actively moving, False otherwise.
		"""
		if self.movingLeft and not self.movingRight:
			self.facing = -1
			return True
		elif not self.movingLeft and self.movingRight:
			self.facing = 1
			return True
		else:
			return False

	def moveSideways( self, dt ):
		# move the player left or right
		self.positionX += self.facing*self.speed*dt
		if self.positionX > self.daisyChain.screenRect.right:
			self.positionX = self.daisyChain.screenRect.left
		elif self.positionX < self.daisyChain.screenRect.left:
			self.positionX = self.daisyChain.screenRect.right
		
	def getPickableFlower( self ):
		"""
		Returns a handle to a Flower if the Player is close enough to a pickable
		flower. Returns False otherwise.
		"""
		# search flowers for a close one
		pickableFlower = None
		closestFlowerDistance = Player.maxFlowerDistance
		for flower in self.floor.flowerIdsToFlowers.values():
			if flower.isPickable():
				playerToFlowerVector = flower.x - self.positionX
				distanceToFlower = math.fabs( playerToFlowerVector )
				if distanceToFlower < closestFlowerDistance:
					closestFlowerDistance = distanceToFlower
					pickableFlower = flower
					try:
						self.playerToFlowerDirection = int( playerToFlowerVector / distanceToFlower ) # -1 or 1
					except ZeroDivisionError:
						pass
		if pickableFlower is not None:
			if closestFlowerDistance > Player.pickingThreshold:
				pickableFlower = None # no valid flowers
		return pickableFlower
	
	def getPullableFlower( self ):
		"""
		Returns a handle to a Flower if the Player is close enough to a pullable
		flower. Returns False otherwise.
		"""
		currentHeight = math.fabs(self.positionY - self.floor.dataBottom)
		if currentHeight < Player.pullingHeight:
			return None # Player isn't high enough
		
		pullableFlower = None
		closestFlowerDistance = Player.maxFlowerDistance
		for flower in self.ceiling.flowerIdsToFlowers.values():
			if flower.isPullable():
				playerToFlowerVector = flower.x - self.positionX
				distanceToFlower = math.fabs( playerToFlowerVector )
				if distanceToFlower < closestFlowerDistance:
					closestFlowerDistance = distanceToFlower
					pullableFlower = flower
					try:
						self.playerToFlowerDirection = int( playerToFlowerVector / distanceToFlower ) # -1 or 1
					except ZeroDivisionError:
						pass
		if pullableFlower is not None:
			if closestFlowerDistance > Player.pullingThreshold:
				pullableFlower = None # no valid flowers
		return pullableFlower

	def renderOtherHalf( self, surface ):
		if self.rect.right > surface.get_rect().right:
			surface.blit( self.image, (self.rect.left - surface.get_rect().right, self.rect.top) )
		elif self.rect.left < surface.get_rect().left:
			surface.blit( self.image, (surface.get_rect().right - (surface.get_rect().left - self.rect.left), self.rect.top) )
